from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ViewDoesNotExist
from django.http import Http404, HttpResponse, HttpResponseServerError, StreamingHttpResponse
from django.http import Http404, HttpResponseNotAllowed
from django.template import TemplateDoesNotExist
from django.views.generic import View

from .converter import ConversionTask
from .decorators import view_function, NotDecoratedError
from .exceptions import BaseRedirectException, InternalRedirectException, RedirectException
from .signals import dmp_signal_pre_process_request, dmp_signal_post_process_request, dmp_signal_internal_redirect_exception, dmp_signal_redirect_exception
from .util import DMP_OPTIONS, get_dmp_instance, log

import inspect
import sys
import threading
from importlib import import_module
from importlib.util import find_spec


# lock to keep get_router() thread safe
rlock = threading.RLock()

# the cache of mini-routers
CACHED_ROUTERS = {}


##############################################################
###   The front controller of all views on the site.
###   urls.py routes everything through this method.

def route_request(request, *args, **kwargs):
    '''
    The main router for all calls coming in to the system.  Patterns in urls.py should call this function.
    '''
    # check to ensure DMP's middleware ran
    # wrap to enable the InternalRedirectExceptions to loop around
    response = None
    while True:
        # an outer try that catches the redirect exceptions
        try:
            # ensure the middleware ran
            if getattr(request, 'dmp', None) is None:
                raise ImproperlyConfigured("Variable request.dmp.function_obj does not exist (check MIDDLEWARE for `django_mako_plus.RequestInitMiddleware`).")

            # output the variables so the programmer can debug where this is routing
            log.info(str(request.dmp))

            # if we had a view not found, raise a 404
            if isinstance(request.dmp.function_obj, RegistryExceptionRouter):
                log.info(request.dmp.function_obj.message(request))
                return request.dmp.function_obj.get_response(request, *args, **kwargs)

            # log the view
            log.info('calling %s', request.dmp.function_obj.message(request))

            # call view function with any args and any remaining kwargs
            response = request.dmp.function_obj.get_response(request, *args, **kwargs)

            # if we didn't get a correct response back, send a 500
            if not isinstance(response, (HttpResponse, StreamingHttpResponse)):
                msg = '%s failed to return an HttpResponse (or the post-signal overwrote it).  Returning 500 error.' % request.dmp.function_obj.message(request)
                log.info(msg)
                return HttpResponseServerError(msg)

            # return the response
            return response

        except InternalRedirectException as ivr:
            # send the signal
            if DMP_OPTIONS['SIGNALS']:
                dmp_signal_internal_redirect_exception.send(sender=sys.modules[__name__], request=request, exc=ivr)
            # resolve to a function
            request.dmp.module = ivr.redirect_module
            request.dmp.function = ivr.redirect_function
            log.info('received an InternalViewRedirect to %s.%s', request.dmp.module, request.dmp.function)
            request.dmp.function_obj = get_router(request.dmp.module, request.dmp.function, verify_decorator=False)
            if isinstance(request.dmp.function_obj, RegistryExceptionRouter):
                log.info('could not fulfill InternalViewRedirect because %s.%s does not exist.', request.dmp.module, request.dmp.function)
            # let it wrap back to the top of the "while True" loop to restart the routing

        except RedirectException as e: # redirect to another page
            if request.dmp.class_obj is None:
                log.info('%s redirected processing to %s', request.dmp.function_obj.message(request), e.redirect_to)
            # send the signal
            if DMP_OPTIONS['SIGNALS']:
                dmp_signal_redirect_exception.send(sender=sys.modules[__name__], request=request, exc=e)
            # send the browser the redirect command
            return e.get_response(request)

    # the code should never get here
    raise Exception("Django-Mako-Plus error: The route_request() function should not have been able to get to this point.  Please notify the owner of the DMP project.  Thanks.")



########################################################
###   Cache of mini routers

def get_router(module_name, function_name, fallback_app=None, fallback_template=None, verify_decorator=True):
    '''
    Gets or creates a mini-router for module_name.function_name.
    Returns one of the four mini-routers defined later in this file.
    If the module or function cannot be found, ViewDoesNotExist is raised.
    '''
    # first check the cache
    key = ( module_name, function_name )
    try:
        return CACHED_ROUTERS[key]
    except KeyError:
        with rlock:
            # try again now that we're locked
            try:
                return CACHED_ROUTERS[key]
            except KeyError:
                func = router_factory(module_name, function_name, fallback_app, fallback_template, verify_decorator)
                if not settings.DEBUG:  # only cache in production mode
                    CACHED_ROUTERS[key] = func
                return func

    # the code should never be able to get here
    raise Exception("Django-Mako-Plus error: registry.get_router() should not have been able to get to this point.  Please notify the owner of the DMP project.  Thanks.")


def router_factory(module_name, function_name, fallback_app=None, fallback_template=None, verify_decorator=True):
    '''
    Factory method to create a view-specific router in the system.
    See the four mini-routers at the end of this file.
    '''
    try:
        # I'm first calling find_spec first here beacuse I don't want import_module in
        # a try/except -- there are lots of reasons that importing can fail, and I just want to
        # know whether the file actually exists.  find_spec raises AttributeError if not found.
        try:
            spec = find_spec(module_name)
        except ValueError:
            spec = None
        if spec is None:
            # no view module, can we call the template directly?
            try:
                return TemplateViewRouter(fallback_app, fallback_template)
            except TemplateDoesNotExist as e:
                raise ViewDoesNotExist('View module {} not found, and fallback template {} could not be loaded ({})'.format(module_name, fallback_template, e))

        # load the module and function
        try:
            module = import_module(module_name)
            func = getattr(module, function_name)
        except ImportError as e:
            raise ViewDoesNotExist('Module "{}" could not be imported: {}'.format(module_name, e))
        except AttributeError as e:
            raise ViewDoesNotExist('Module "{}" found successfully, but "{}" was not found: {}'.format(module_name, function_name, e))

        # get the converter from the decorator kwargs, if there is one
        try:
            decorator_kwargs = view_function.get_kwargs(func)[0]
        except NotDecoratedError:
            decorator_kwargs = {}

        # class-based view?
        if inspect.isclass(func) and issubclass(func, View):
            return ClassBasedRouter(module, func(), decorator_kwargs)  # func() to instantiate because func is class (not instance)

        # a view function?
        if verify_decorator and not view_function.is_decorated(func):
            raise ViewDoesNotExist("View {}.{} was found successfully, but it must be decorated with @view_function or be a subclass of django.views.generic.View.".format(module_name, function_name))
        return ViewFunctionRouter(module, func, decorator_kwargs)

    except ViewDoesNotExist as vdne:
        return RegistryExceptionRouter(vdne)




########################################################################################
###   Router classes for single endpoints.  When a view is first accessed, one
###   of these "mini" routers is created for it and cached for future calls.


class ViewFunctionRouter(object):
    '''Router for view functions and class-based methods'''
    def __init__(self, mod, func, decorator_kwargs):
        self.module = mod
        self.function = func
        self.decorator_kwargs = decorator_kwargs
        self.signature = inspect.signature(func)
        # map the type hints
        param_types = getattr(func, '__annotations__', {})  # not using typing.get_type_hints because it adds Optional() to None defaults, and we don't need to follow mro here
        # inspect the parameters of the view function
        params = []
        for i, p in enumerate(self.signature.parameters.values()):
            params.append(ViewParameter(
                name=p.name,
                position=i,
                kind=p.kind,
                type=param_types.get(p.name) or inspect.Parameter.empty,
                default=p.default,
            ))
        self.parameters = tuple(params)


    def get_response(self, *args, **kwargs):
        '''Converts urlparams, calls the view function, returns the response'''
        # leaving request inside *args (or kwargs) so we can convert it like anything else (and parameter indices aren't off by one)
        request = kwargs.get('request', args[0])
        ctask = ConversionTask(request, self.module, self.function, self.decorator_kwargs)
        args = list(args)
        urlparam_i = 0
        # add urlparams into the arguments and convert the values
        for parameter_i, parameter in enumerate(self.parameters):
            try:
                # skip *args, **kwargs
                if parameter.kind is inspect.Parameter.VAR_POSITIONAL or parameter.kind is inspect.Parameter.VAR_KEYWORD:
                    continue
                # value in kwargs?
                elif parameter.name in kwargs:
                    kwargs[parameter.name] = ctask.converter(kwargs[parameter.name], parameter, ctask)
                # value in args?
                elif parameter_i < len(args):
                    args[parameter_i] = ctask.converter(args[parameter_i], parameter, ctask)
                # urlparam value?
                elif urlparam_i < len(request.dmp.urlparams):
                    kwargs[parameter.name] = ctask.converter(request.dmp.urlparams[urlparam_i], parameter, ctask)
                    urlparam_i += 1
                # can we assign a default value?
                elif parameter.default is not inspect.Parameter.empty:
                    kwargs[parameter.name] = ctask.converter(parameter.default, parameter, ctask)
                # fallback is None
                else:
                    kwargs[parameter.name] = ctask.converter(None, parameter, ctask)

            except BaseRedirectException as e:
                log.info('Redirect exception raised during conversion of parameter %s (%s): %s', parameter.position, parameter.name, e)
                raise
            except Http404 as e:
                log.info('Raising Http404 because exception raised during conversion of parameter %s (%s): %s', parameter.position, parameter.name, e, exc_info=e)
                raise
            except Exception as e:
                log.info('Raising Http404 because exception raised during conversion of parameter %s (%s): %s', parameter.position, parameter.name, e, exc_info=e)
                raise Http404('Invalid parameter specified in the url')

        # send the pre-signal
        if DMP_OPTIONS['SIGNALS']:
            for receiver, ret_response in dmp_signal_pre_process_request.send(sender=sys.modules[__name__], request=request, view_args=args, view_kwargs=kwargs):
                if isinstance(ret_response, (HttpResponse, StreamingHttpResponse)):
                    return ret_response

        # call the view!
        response = self.function(*args, **kwargs)

        # send the post-signal
        if DMP_OPTIONS['SIGNALS']:
            for receiver, ret_response in dmp_signal_post_process_request.send(sender=sys.modules[__name__], request=request, response=response, view_args=args, view_kwargs=kwargs):
                if ret_response is not None:
                    response = ret_response # sets it to the last non-None in the signal receiver chain

        # return the response
        return response


    def message(self, request):
        return 'view function {}.{}'.format(request.dmp.module, request.dmp.function)



class ClassBasedRouter(object):
    '''Router for class-based views.'''
    def __init__(self, module, instance, decorator_kwargs):
        self.instance = instance
        self.endpoints = {}
        for mthd_name in instance.http_method_names:  # get parameters from the first http-based method (get, post, etc.)
            func = getattr(instance, mthd_name, None)
            if func is not None:
                self.endpoints[mthd_name] = ViewFunctionRouter(module, func, decorator_kwargs)


    def get_response(self, request, *args, **kwargs):
        endpoint = self.endpoints.get(request.method.lower())
        if endpoint is not None:
            return endpoint.get_response(request, **kwargs)
        log.info('Method Not Allowed (%s): %s', request.method, request.path, extra={'status_code': 405, 'request': request})
        return HttpResponseNotAllowed([ e.upper() for e in self.endpoints.keys() ])


    @property
    def name(self):
        return self.instance.__class__.__qualname__


    def message(self, request):
        return 'class-based view function {}.{}.{}'.format(request.dmp.module, self.name, request.dmp.function)



class TemplateViewRouter(object):
    '''Router for direct templates (used whe a view.py file doesn't exist but the .html does)'''
    def __init__(self, app_name, template_name):
        # not keeping the actual template objects because we need to get from the loader each time (Mako has its own cache)
        self.app_name = app_name
        self.template_name = template_name
        # check the template by loading it
        get_dmp_instance().get_template_loader(self.app_name).get_template(self.template_name)


    def get_response(self, request, *args, **kwargs):
        template = get_dmp_instance().get_template_loader(self.app_name).get_template(self.template_name)
        return template.render_to_response(request=request, context=kwargs)


    def message(self, request):
        return 'template {} (view function {}.{} not found)'.format(self.template_name, request.dmp.module, request.dmp.function)



class RegistryExceptionRouter(object):
    '''Router for a registry exception (i.e. view not found).'''
    def __init__(self, exc):
        self.exc = exc


    def get_response(self, request, *args, **kwargs):
        raise Http404(str(self.exc))


    def message(self, request):
        return str(self.exc)



#####################################
###  ViewParameter

class ViewParameter(object):
    '''
    A data class that represents a view parameter on a view function.
    An instance of this class is created for each parameter in a view function
    (except the initial request object argument).
    '''
    def __init__(self, name, position, kind, type, default):
        '''
        name:      The name of the parameter.
        position:  The position of this parameter.
        kind:      The kind of argument (positional, keyword, etc.). See inspect module.
        type:      The expected type of this parameter.  Converters use this type to
                   convert urlparam strings to the right type.
        default:   Any default value, specified in function type hints.  If no default is
                   specified in the function, this is `inspect.Parameter.empty`.
        '''
        self.name = name
        self.position = position
        self.kind = kind
        self.type = type
        self.default = default

    def __str__(self):
        return 'ViewParameter: name={}, type={}, default={}'.format(
            self.name,
            self.type.__qualname__ if self.type is not None else '<not specified>',
            self.default,
        )






from django.core import management
from django.core.management.base import BaseCommand
import django_mako_plus
import os, os.path


class Command(BaseCommand):
    args = '<name>'
    help = 'Creates a new Django-Mako-Plus app.'
    can_import_settings = True
    missing_args_message = "You must provide an application name."


    def add_arguments(self, parser):
        # required argument for the app name
        parser.add_argument('appname', type=str, help='the name of the new app')


    def handle(self, *args, **options):
        # figure out the DMP app_template directory
        url = 'http://cdn.rawgit.com/doconix/django-mako-plus/master/app_template.zip'
        if django_mako_plus.__file__:
            dmp_dir = os.path.dirname(django_mako_plus.__file__)
            template_dir = os.path.join(dmp_dir, 'app_template')
            if os.path.exists(template_dir):
                url = template_dir

        # redirect
        self.stdout.write('Redirecting to startapp --template={} --extension=py,htm,html {}'.format(url, options['appname']))
        management.call_command('startapp', options['appname'], template=url, extension='py,htm,html')




from django.core.management.commands.makemessages import Command as MakeMessagesCommand

from django_mako_plus.registry import get_dmp_apps
from django_mako_plus.convenience import get_template_for_path

import os
import os.path


class Command(MakeMessagesCommand):
    help = "Makes messages for Mako templates.  The native makemessages in Django doesn't understand Mako files, so this command compiles all your templates so it can find the compiled .py versions of the templates."
    SEARCH_DIRS = [
        os.path.join('{app_path}', 'templates')
    ]

    def message(self, msg, level=1):
        '''Print a message to the console'''
        # verbosity=1 is the default if not specified in the options
        if self.options['verbosity'] >= level:
            print(msg)


    def add_arguments(self, parser):
        # must call super because not directly subclassing BaseCommand
        super(Command, self).add_arguments(parser)
        # add our additional command
        parser.add_argument(
            '--verbose',
            action='store_true',
            dest='verbose',
            default=False,
            help='Set verbosity to level 3 (see --verbosity).'
        )
        parser.add_argument(
            '--quiet',
            action='store_true',
            dest='quiet',
            default=False,
            help='Set verbosity to level 0, which silences all messages (see --verbosity).'
        )
        parser.add_argument(
            '--extra-gettext-option',
            default=[],
            dest='extra_gettext_option',
            action='append',
            help="Add an additional option to be passed to gettext. Ex: --extra-gettext-option='--keyword=mytrans'"
        )


    def handle(self, *args, **options):
        self.options = options
        if self.options['verbose']:
            self.options['verbosity'] = 3
        if self.options['quiet']:
            self.options['verbosity'] = 0

        # go through each dmp_enabled app and compile its mako templates
        for app_config in get_dmp_apps():
            self.compile_mako_files(app_config)

        # add any extra xgettext_options (the regular makemessages doesn't do this, and I need to include other aliases like _(), _z(), etc.
        for opt in options.get('extra_gettext_option', []):
           self.xgettext_options.append(opt)

        # call the superclass command
        return super(Command, self).handle(*args, **options)


    def compile_mako_files(self, app_config):
        '''Compiles the Mako templates within the apps of this system'''
        # go through the files in the templates, scripts, and styles directories
        for subdir_name in self.SEARCH_DIRS:
            subdir = subdir_name.format(
                app_path=app_config.path,
                app_name=app_config.name,
            )
            if os.path.exists(subdir):
                for filename in os.listdir(subdir):
                    _, ext = os.path.splitext(filename)
                    if ext.lower() in ( '.htm', '.html', '.mako' ):
                        # create the template object, which creates the compiled .py file
                        filepath = os.path.join(subdir, filename)
                        self.message(filepath, 2)
                        get_template_for_path(filepath)



from ..command import run_command
from ..util import merge_dicts, flatten
from .base import BaseProvider

import os
import os.path
import shutil



class CompileProvider(BaseProvider):
    '''
    Runs a command, such as compiling *.scss or *.less, when an output file
    timestamp is older than the source file. In production mode, this check
    is done only once (the first time a template is run) per server start.

    Subclasses can override `source` and `target`, OR just override `needs_compile`.
    '''
    default_options = merge_dicts(BaseProvider.default_options, {
        'group': 'styles',
    })
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.needs_compile:
            run_command(*self.build_command())

    @property
    def source(self):
        return '/path/to/source.file'

    @property
    def target(self):
        return '/path/to/compiled/target.file'

    def build_command(self):
        '''Returns the command to run, as a list/tuple (see subprocess module)'''
        return ( 'echo', 'Subclasses should override this command.' )

    @property
    def needs_compile(self):
        try:
            source_mtime = os.stat(self.source).st_mtime
        except OSError:  # no source for this template, so just return
            return False
        try:
            target_mtime = os.stat(self.target).st_mtime
        except OSError: # target doesn't exist, so compile
            return True
        # both source and target exist, so compile if source newer
        return source_mtime > target_mtime


class CompileScssProvider(CompileProvider):
    '''Specialized CompileProvider that contains settings for *.scss files.'''
    @property
    def source(self):
        # this inner
        return os.path.join(*flatten(self.app_config.path, 'styles', self.subdir_parts[1:], self.template_name + '.scss'))

    @property
    def target(self):
        return os.path.join(*flatten(self.app_config.path, 'styles', self.subdir_parts[1:], self.template_name + '.css'))

    def build_command(self):
        return (
            shutil.which('scss'),
            '--load-path=.',
            '--unix-newlines',
            self.source,
            self.target,
         )


class CompileLessProvider(CompileProvider):
    '''Specialized CompileProvider that contains settings for *.less files.'''
    @property
    def source(self):
        return os.path.join(*flatten(self.app_config.path, 'styles', self.subdir_parts[1:], self.template_name + '.less'))

    @property
    def target(self):
        return os.path.join(*flatten(self.app_config.path, 'styles', self.subdir_parts[1:], self.template_name + '.css'))

    def build_command(self):
        return (
            shutil.which('lessc'),
            '--source-map',
            self.source,
            self.target,
         )





from django.apps import apps as django_apps
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from mako.exceptions import MakoException
from django_mako_plus.registry import get_dmp_apps, ensure_dmp_app
from django_mako_plus.provider import create_mako_context
from django_mako_plus.provider.runner import ProviderRun, create_factories
from django_mako_plus.util import get_dmp_instance, split_app, DMP_OPTIONS

import glob
import os, os.path, shutil
import json
from collections import OrderedDict



class Command(BaseCommand):
    args = ''
    help = 'Removes compiled template cache folders in your DMP-enabled app directories.'
    can_import_settings = True


    def add_arguments(self, parser):
        parser.add_argument(
            '--verbose',
            action='store_true',
            dest='verbose',
            default=False,
            help='Set verbosity to level 3 (see --verbosity).'
        )
        parser.add_argument(
            '--quiet',
            action='store_true',
            dest='quiet',
            default=False,
            help='Set verbosity to level 0, which silences all messages (see --verbosity).'
        )
        parser.add_argument(
            '--overwrite',
            action='store_true',
            dest='overwrite',
            default=False,
            help='Overwrite existing __entry__.js if necessary.'
        )
        parser.add_argument(
            '--single',
            type=str,
            metavar='FILENAME',
            help='Instead of per-app entry files, create a single file that includes the JS for all listed apps.'
        )
        parser.add_argument(
            'appname',
            type=str,
            nargs='*',
            help='The name of one or more DMP apps. If omitted, all DMP apps are processed.'
        )


    def message(self, msg, level=1):
        '''Print a message to the console'''
        # verbosity=1 is the default if not specified in the options
        if self.options['verbosity'] >= level:
            print(msg)


    def handle(self, *args, **options):
        # save the options for later
        self.options = options
        if self.options['verbose']:
            self.options['verbosity'] = 3
        if self.options['quiet']:
            self.options['verbosity'] = 0
        self.factories = create_factories('WEBPACK_PROVIDERS')

        # ensure we have a base directory
        try:
            if not os.path.isdir(os.path.abspath(settings.BASE_DIR)):
                raise CommandError('Your settings.py BASE_DIR setting is not a valid directory.  Please check your settings.py file for the BASE_DIR variable.')
        except AttributeError as e:
            print(e)
            raise CommandError('Your settings.py file is missing the BASE_DIR setting.')

        # the apps to process
        apps = []
        for appname in options['appname']:
            ensure_dmp_app(appname)
            apps.append(django_apps.get_app_config(appname))
        if len(apps) == 0:
            apps = get_dmp_apps()

        # main runner for per-app files
        if self.options['single'] is None:
            for app in apps:
                self.message('Searching `{}` app...'.format(app.name))
                filename = os.path.join(app.path, 'scripts', '__entry__.js')
                self.create_entry_file(filename, self.generate_script_map(app), [ app ])

        # main runner for one sitewide file
        else:
            script_map = {}
            for app in apps:
                self.message('Searching `{}` app...'.format(app.name))
                script_map.update(self.generate_script_map(app))
            self.create_entry_file(self.options['single'], script_map, apps)


    def create_entry_file(self, filename, script_map, apps):
        '''Creates an entry file for the given script map'''
        def in_apps(s):
            for app in apps:
                last = None
                path = os.path.dirname(s)
                while last != path:
                    if os.path.samefile(app.path, path):
                        return True
                    last = path
                    path = os.path.dirname(last)
            return False

        filedir = os.path.dirname(filename)
        if os.path.exists(filename):
            if self.options['overwrite']:
                os.remove(filename)
            else:
                raise ValueError('Refusing to destroy existing file: {} (use --overwrite option or remove the file)'.format(filename))
        # create the lines of the entry file
        lines = []
        for page, scripts in script_map.items():
            require = []
            for s in scripts:
                if in_apps(os.path.abspath(s)):
                    require.append('require("./{}")'.format(os.path.relpath(s, filedir)))
            if len(require) > 0:
                lines.append('DMP_CONTEXT.appBundles["{}"] = () => {{ {}; }};'.format(page, '; '.join(require)))
        # if we created at least one line, write the entry file
        if len(lines) > 0:
            self.message('Creating {}'.format(os.path.relpath(filename, settings.BASE_DIR)))
            with open(filename, 'w') as fout:
                fout.write('(context => {\n')
                for line in lines:
                    fout.write('    {}\n'.format(line))
                fout.write('})(DMP_CONTEXT.get());\n')


    def generate_script_map(self, config):
        '''
        Maps templates in this app to their scripts.  This function iterates through
        app/templates/* to find the templates in this app.  Returns the following
        dictionary with paths relative to BASE_DIR:

        {
            'app/template1': [ 'scripts/template1.js', 'scripts/supertemplate1.js' ],
            'app/template2': [ 'scripts/template2.js', 'scripts/supertemplate2.js', 'scripts/supersuper2.js' ],
            ...
        }

        Any files or subdirectories starting with double-underscores (e.g. __dmpcache__) are skipped.
        '''
        script_map = OrderedDict()
        template_root = os.path.join(config.path, 'templates')
        def recurse(folder):
            for filename in os.listdir(os.path.join(template_root, folder)):
                if filename.startswith('__'):
                    continue
                filerel = os.path.join(folder, filename)
                filepath = os.path.join(os.path.join(template_root, filerel))
                if os.path.isdir(filepath):
                    recurse(filerel)

                elif os.path.isfile(filepath):
                    scripts = self.template_scripts(config, filerel)
                    key = '{}/{}'.format(config.name, os.path.splitext(filerel)[0])
                    if len(scripts) > 0:
                        script_map[key] = scripts
                        self.message('\t{}: {}'.format(key, scripts), 3)
                    else:
                        self.message('\t{}: none found'.format(key), 3)
        recurse('')
        return script_map


    def template_scripts(self, config, template_name):
        '''
        Returns a list of scripts used by the given template object AND its ancestors.

        This runs a ProviderRun on the given template (as if it were being displayed).
        This allows the WEBPACK_PROVIDERS to provide the JS files to us.

        For this algorithm to work, the providers must extend static_links.LinkProvider
        because that class creates the 'enabled' list of provider instances, and each
        provider instance has a url.
        '''
        template_obj = get_dmp_instance().get_template_loader(config, create=True).get_mako_template(template_name, force=True)
        mako_context = create_mako_context(template_obj)
        inner_run = ProviderRun(mako_context['self'], factories=self.factories)
        inner_run.run()
        scripts = []
        for data in inner_run.column_data:
            for provider in data.get('enabled', []):
                if getattr(provider, 'url') is not None:
                    url = provider.url.split('?', 1)[0]
                    scripts.append(os.path.relpath(url, settings.BASE_DIR))
        return scripts



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from .base import BaseProvider
from ..template import template_inheritance
from ..util import DMP_OPTIONS, split_app, merge_dicts
from ..uid import wuid

import io
from collections import namedtuple
import warnings


##################################################
###   Static File Provider Factory


def init_providers():
    '''Called when the DMP template engine is created by Django'''
    DMP_OPTIONS['RUNTIME_PROVIDER_FACTS'] = create_factories()


def create_factories(key='CONTENT_PROVIDERS'):
    '''Called from here as well as dmp_webpack.py'''
    factories = []
    for index, provider_def in enumerate(DMP_OPTIONS[key]):
        fac = ProviderFactory(provider_def, '_django_mako_plus_providers_{}_{}_'.format(key, index))
        if fac.options['enabled']:  # providers can be disabled globally in settings (to run on debug or prod only, for example)
            factories.append(fac)
    return factories


class ProviderFactory(object):
    '''Creator for a given Provider definition in settings.py.'''
    def __init__(self, provider_def, cache_key):
        self.cache_key = cache_key
        self.options = {}
        try:
            self.provider_class = provider_def['provider']
        except KeyError:
            raise ImproperlyConfigured('The Django Mako Plus template OPTIONS were not set up correctly in settings.py; a CONTENT_PROVIDERS item is missing `provider`.')
        if isinstance(self.provider_class, str):
            self.provider_class = import_string(self.provider_class)
        if not issubclass(self.provider_class, BaseProvider):
            raise ImproperlyConfigured('The Django Mako Plus template OPTIONS were not set up correctly in settings.py; The `provider` value must be a subclass of django_mako_plus.BaseProvider.')
        self.options = merge_dicts(self.provider_class.default_options, provider_def)

    def instance_for_template(self, template):
        '''Returns a provider instance for the given template'''
        # Mako already caches template objects, so I'm attaching the instance to the template for speed during production
        try:
            return getattr(template, self.cache_key)
        except AttributeError:
            pass

        # create and cache (if in prod mode)
        app_config, template_file = split_app(template.filename)
        if app_config is None:
            raise ImproperlyConfigured("Could not determine the app for template {}. Is this template's app registered as a DMP app?".format(template.filename))
        instance = self.provider_class(app_config, template_file, self.options)
        if not settings.DEBUG:
            setattr(template, self.cache_key, instance)
        return instance



####################################################
###   Main runner


class ProviderRun(object):
    '''A run through the providers for tself and its ancestors'''
    def __init__(self, tself, version_id=None, group=None, factories=None):
        '''
        tself:      `self` object from a Mako template (available during rendering).
        version_id: hash/unique number to place on links
        group:      provider group to include (defaults to all groups if None)
        factories:  list of provider factories to run on each template (defaults to settings.py if None)
        '''
        self.uid = wuid()           # a unique id for this run
        self.request = tself.context.get('request')
        self.context = tself.context
        if version_id is not None:
            warnings.warn('The `version_id` parameter in links() is deprecated in favor of automatic file hashes.')
        self.buffer = io.StringIO()
        # Create a table of providers for each template in the ancestry:
        #
        #     base.htm,      [ JsLinkProvider1, CssLinkProvider1, ... ]
        #        |
        #     app_base.htm,  [ JsLinkProvider2, CssLinkProvider2, ... ]
        #        |
        #     index.html,    [ JsLinkProvider3, CssLinkProvider3, ... ]
        factories = [
            pf
            for pf in (factories if factories is not None else DMP_OPTIONS['RUNTIME_PROVIDER_FACTS'])
            if group is None or group == pf.options['group']
        ]
        self.template_providers = [
            [ pf.instance_for_template(template) for pf in factories ]
            for template in reversed(list(template_inheritance(tself)))
        ]
        # Column-specific data dictionaries are maintained as the template providers run
        # (one per provider column).  These allow the provider instances of a given
        # column to share data if needed.
        #
        #      column_data = [ { col 1 }      , { col 2 }      , ... ]
        self.column_data = [ {} for pf in factories ]


    def run(self):
        '''Performs the run through the templates and their providers'''
        # start() on tself (the last template in the list)
        for providers in self.template_providers[-1:]:
            for provider, data in zip(providers, self.column_data):
                provider.start(self, data)
        # provide() on the all provider lists in the chain
        for providers in self.template_providers:
            for provider, data in zip(providers, self.column_data):
                provider.provide(self, data)
        # finish() on tself (the last template in the list)
        for providers in self.template_providers[-1:]:
            for provider, data in zip(providers, self.column_data):
                provider.finish(self, data)


    def write(self, content):
        '''Provider instances use this to write to the buffer'''
        self.buffer.write(content)
        if settings.DEBUG:
            self.buffer.write('\n')


    def getvalue(self):
        '''Returns the buffer string'''
        return self.buffer.getvalue()



from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

import base64
import os, os.path
import collections
import zlib


# this is populated with the dictionary of options in engine.py when
# Django initializes the template engine
DMP_OPTIONS = {}

# the key for the DMP singleton engine instance (stored in DMP_OPTIONS by engine.py)
DMP_INSTANCE_KEY = 'django_mako_plus_instance'

# set up the logger
import logging
log = logging.getLogger('django_mako_plus')

################################################################
###   DMP instance and apps

def get_dmp_instance():
    '''
    Retrieves the DMP template engine instance.
    '''
    try:
        return DMP_OPTIONS[DMP_INSTANCE_KEY]
    except KeyError:
        raise ImproperlyConfigured('The Django Mako Plus template engine did not initialize correctly.  Check your logs for previous errors that may have caused initialization to fail, and check that DMP is set correctly in settings.py.')




################################################################
###   Special type of list used for url params

class URLParamList(list):
    '''
    A simple extension to Python's list that returns '' for indices that don't exist.
    For example, if the object is ['a', 'b'] and you call obj[5], it will return ''
    rather than throwing an IndexError.  This makes dealing with url parameters
    simpler since you don't have to check the length of the list.
    '''
    def __getitem__(self, idx):
        '''Returns the element at idx, or '' if idx is beyond the length of the list'''
        return self.get(idx, '')

    def get(self, idx, default=''):
        '''Returns the element at idx, or default if idx is beyond the length of the list'''
        # if the index is beyond the length of the list, return ''
        if isinstance(idx, int) and (idx >= len(self) or idx < -1 * len(self)):
            return default
        # else do the regular list function (for int, slice types, etc.)
        return super().__getitem__(idx)


##########################################
###   Utilities

def merge_dicts(*dicts):
    '''
    Merges an arbitrary number of dicts, starting
    with the first argument and updating through the
    last argument (last dict wins on conflicting keys).
    '''
    merged = {}
    for d in dicts:
        merged.update(d)
    return merged


def flatten(*args):
    '''Generator that recursively flattens embedded lists, tuples, etc.'''
    for arg in args:
        if isinstance(arg, collections.Iterable) and not isinstance(arg, (str, bytes)):
            yield from flatten(*arg)
        else:
            yield arg


def split_app(path):
    '''
    Splits a path on the app, returning (app config, relative path within app).

    This function uses os.path.samefile to split even if a path is specified differently.
    The drawback is significantly reduced speed because of a double loop, so use with care.
    ProviderRun uses this, but only during the first run of a template (cached after that).
    '''
    from django_mako_plus.registry import get_dmp_apps
    configs = list(get_dmp_apps())
    def get_config(app_dir):
        if os.path.exists(app_dir):
            for config in configs:
                if os.path.samefile(config.path, app_dir):
                    return config
        return None
    parts = os.path.normpath(path).split(os.path.sep)
    for i in range(len(parts), 0, -1):
        app_config = get_config(os.path.sep.join(parts[:i]))
        if app_config is not None:
            return app_config, os.path.sep.join(parts[i:])
    return None, path


def crc32(filename):
    '''
    Calculates the CRC checksum for a file.
    Using CRC32 because security isn't the issue and don't need perfect noncollisions.
    We just need to know if a file has changed.

    On my machine, crc32 was 20 times faster than any hashlib algorithm,
    including blake and md5 algorithms.
    '''
    result = 0
    with open(filename, 'rb') as fin:
        while True:
            chunk = fin.read(48)
            if len(chunk) == 0:
                break
            result = zlib.crc32(chunk, result)
    return result


EMPTY = object()
def getdefaultattr(obj, name, default=None, factory=EMPTY):
    '''
    Gets the given attribute from the object,
    creating it with a default or by calling
    a factory if needed.
    '''
    try:
        return getattr(obj, name)
    except AttributeError:
        pass
    val = factory() if factory is not EMPTY else None
    setattr(obj, name, val)
    return val



from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from django_mako_plus.registry import get_dmp_apps

import glob
import os
import os.path




class Command(BaseCommand):
    args = ''
    help = 'Removes orphaned *.css, *.css.map, *.cssm, and *.cssm.map files from your DMP styles/ folders that no longer have companion *.scss files.'
    can_import_settings = True


    def add_arguments(self, parser):
        parser.add_argument(
            '--trial-run',
            action='store_true',
            dest='trial_run',
            default=False,
            help='Display the files that would be removed without actually removing them.'
        )
        parser.add_argument(
            '--verbose',
            action='store_true',
            dest='verbose',
            default=False,
            help='Set verbosity to level 3 (see --verbosity).'
        )
        parser.add_argument(
            '--quiet',
            action='store_true',
            dest='quiet',
            default=False,
            help='Set verbosity to level 0, which silences all messages (see --verbosity).'
        )
        parser.add_argument(
            '--directory',
            action='store',
            dest='directory',
            default='',
            help='Check the given directory instead of searching through DMP app styles/ folders.'
        )
        parser.add_argument(
            '--recursive',
            action='store_true',
            dest='recursive',
            default=False,
            help="Recurse to subdirectories in each search folder.  This option searches subdirectories of each app's styles/ folder, or when specified with --directory, searches subdirectories of the given directory."
        )


    def handle(self, *args, **options):
        # save the options for later
        self.options = options
        if self.options['verbose']:
            self.options['verbosity'] = 3
        if self.options['quiet']:
            self.options['verbosity'] = 0
        if self.options['trial_run']:
            self.message("Trial run: dmp_sass_cleanup would have deleted the following files:", level=1)

        # ensure we have a base directory
        try:
            if not os.path.isdir(os.path.abspath(settings.BASE_DIR)):
                raise CommandError('Your settings.py BASE_DIR setting is not a valid directory.  Please check your settings.py file for the BASE_DIR variable.')
        except AttributeError as e:
            print(e)
            raise CommandError('Your settings.py file is missing the BASE_DIR setting.')

        # create a list of directories to check (default to each app's styles/)
        styles_directories = [ os.path.join(config.path, 'styles') for config in get_dmp_apps() ]
        if self.options['directory']:
            if not os.path.isdir(self.options['directory']):
                raise CommandError('The specified directory does not exist: {}'.format(self.options['directory']))
            styles_directories = [ self.options['directory'] ]

        # if we are in recursive mode, add all subdirectories of the search folders
        if self.options['recursive']:
            styles_directories = [ d[0] for styles_dir in styles_directories for d in os.walk(styles_dir) ]

        # check the directories
        for styles_dir in styles_directories:
            self.message('Checking directory {}'.format(pretty_relpath(styles_dir, settings.BASE_DIR)), level=2)
            self.clean_styles(styles_dir)


    def message(self, msg, level):
        '''Print a message to the console'''
        # verbosity=1 is the default if not specified in the options
        if self.options['verbosity'] >= level:
            print(msg)


    def clean_styles(self, styles_dir):
        '''Removes orphaned .css and .css.map files in the given styles directory.'''
        # go through the files in this directory
        for ext in ( 'css', 'cssm' ):
            for csspath in glob.glob(os.path.join(styles_dir, '*.{}'.format(ext))):
                relcss = pretty_relpath(csspath, settings.BASE_DIR)
                scsspath = '{}.s{}'.format(os.path.splitext(csspath)[0], ext)
                mappath = '{}.{}.map'.format(os.path.splitext(csspath)[0], ext)

                # if the .scss file exists, it's not orphaned
                if os.path.exists(scsspath):
                    self.message('Skipping {} because its .scss file still exists.'.format(relcss), level=3)
                    continue

                # if the css has no sourceMappingURL line, it's not a generated sass file
                smu = '/*# sourceMappingURL={}.map */'.format(os.path.split(csspath)[1])
                if not file_has_line(csspath, smu):
                    self.message('Skipping {} because it is not a generated Sass file (no "{}").'.format(relcss, smu), level=3)
                    continue

                # if we get here, the css should be removed
                self.message('Removing {} because it is a generated Sass file, but {} no longer exists.'.format(relcss, pretty_relpath(scsspath, settings.BASE_DIR)), level=1)
                if not self.options['trial_run']:
                    os.remove(csspath)

                # see if the .css.map file also needs removing
                if os.path.exists(mappath):
                    self.message('Removing {} also.'.format(pretty_relpath(mappath, settings.BASE_DIR)), level=1)
                    if not self.options['trial_run']:
                        os.remove(mappath)




#####################################################
###   Utility functions


def file_has_line(filepath, st):
    '''
    Returns true if the given file has a line that starts with the given st
    '''
    with open(filepath) as f:
        for line in f:
            if line.lstrip().startswith(st):
                return True
    return False


def pretty_relpath(path, start):
    '''
    Returns a relative path, but only if it doesn't start with a non-pretty parent directory ".."
    '''
    relpath = os.path.relpath(path, start)
    if relpath.startswith('..'):
        return path
    return relpath



from django.conf import settings
from django.utils.module_loading import import_string

from ..util import merge_dicts, flatten, log, crc32, getdefaultattr
from ..version import __version__
from .base import BaseProvider

import logging
import os
import os.path
import json
import glob
from collections import namedtuple


#####################################################
###   Abstract Link Provider

class LinkProvider(BaseProvider):
    '''
    Renders links like <link> and <script> based on the name of the template
    and supertemplates.
    '''


    # the default options are for CSS files
    default_options = merge_dicts(BaseProvider.default_options, {
        # subclasses should override the group
        'group': 'scripts',
        # string for the absolute path to the filename that should be linked (if it exists)
        # if this is a function/lambda, DMP will run the function to get the string path
        'filename': 'app/subdir/template.ext',
        # how to process duplicate urls found across this request
        'skip_duplicates': False,
    })
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        filepath = self.options['filename']
        if callable(filepath):  # if a function, call it to get the path
            filepath = filepath(self)
        self.filename = os.path.relpath(filepath, settings.BASE_DIR).replace('\\', '/')
        if log.isEnabledFor(logging.DEBUG):
            log.debug('[%s] %s searching for %s', self.template_file, self.__class__.__qualname__, self.filename)
        # file time and version hash
        try:
            self.mtime = int(os.stat(filepath).st_mtime)
            if log.isEnabledFor(logging.INFO):
                log.info('[%s] found %s', self.template_file, self.filename)
            # version_id combines current time and the CRC32 checksum of file bytes
            self.version_id = (self.mtime << 32) | crc32(filepath)
            self.url = '{}?v={:x}'.format(
                self.filename.replace('\\', '/'),
                self.version_id,
            )
        except FileNotFoundError:
            self.mtime = 0
            self.version_id = 0
            self.url = None

    def start(self, provider_run, data):
        # add a set to the request (fallback to provider_run if request is None) for skipping duplicates
        if self.options['skip_duplicates']:
            data['seen'] = getdefaultattr(
                provider_run.request.dmp if provider_run.request is not None else provider_run,
                '_LinkProvider_Filename_Cache_',
                factory=set,
            )
        # enabled providers in the chain go here
        data['enabled'] = []

    def provide(self, provider_run, data):
        # delaying printing of tag to finish() because the JsContextProvider delays and this must go after it

        # short circuit if the file for this provider doesn't exist
        if self.url is None:
            return
        # short circut if we're skipping duplicates and we've already seen this one
        if self.options['skip_duplicates']:
            if self.filename in data['seen']:
                return
            data['seen'].add(self.filename)
        # if we get here, this provider is enabled, so add it to the list
        data['enabled'].append(self)

    def finish(self, provider_run, data):
        for provider in data['enabled']:
            provider_run.write(provider.create_link(provider_run))

    def create_link(self, provider_run, url):
        raise NotImplementedError('Subclass did not implement `create_link()`')


#############################################
###   Css and Js Link Providers


class CssLinkProvider(LinkProvider):
    '''Generates a CSS <link>'''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'styles',
        'filename': lambda pr: os.path.join(*flatten(pr.app_config.path, 'styles', pr.subdir_parts[1:], pr.template_name + '.css')),
        'skip_duplicates': True,
    })

    def create_link(self, provider_run):
        '''Creates a link to the given URL'''
        return '<link data-context="{contextid}" rel="stylesheet" type="text/css" href="{static}{url}" />'.format(
            contextid=provider_run.uid,
            static=settings.STATIC_URL,
            url=self.url,
        )


class JsLinkProvider(LinkProvider):
    '''Generates a JS <script>.'''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'scripts',
        'filename': lambda pr: os.path.join(*flatten(pr.app_config.path, 'scripts', pr.subdir_parts[1:], pr.template_name + '.js')),
        'async': False,
        'skip_duplicates': False,  # the JS may need to run each time it is included
    })

    def create_link(self, provider_run):
        '''Creates a link to the given URL'''
        return '<script data-context="{contextid}" src="{static}{url}"{async}></script>'.format(
            contextid=provider_run.uid,
            static=settings.STATIC_URL,
            url=self.url,
            async=' async="async"' if self.options['async'] else '',
        )



###################################
###  JS Context Provider

class JsContextProvider(BaseProvider):
    '''
    Adds all js_context() variables to DMP_CONTEXT.
    This should be listed before JsLinkProvider so the
    context variables are available during <script> runs.
    '''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'scripts',
        'encoder': 'django.core.serializers.json.DjangoJSONEncoder',
    })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encoder = import_string(self.options['encoder'])
        self.template = "{}/{}".format(self.app_config.name, self.template_name)

    def start(self, provider_run, data):
        data['templates'] = []

    def provide(self, provider_run, data):
        data['templates'].append(self.template)

    def finish(self, provider_run, data):
        if len(data['templates']) == 0:
            return
        context_data = {
            jscontext('__router__'): {
                'template': self.template,
                'app': provider_run.request.dmp.app if provider_run.request is not None else None,
                'page': provider_run.request.dmp.page if provider_run.request is not None else None,
            },
        }
        for k in provider_run.context.kwargs:
            if isinstance(k, jscontext):
                value = provider_run.context[k]
                context_data[k] = value.__jscontext__() if hasattr(value, '__jscontext__') else value
        provider_run.write('<script>')
        provider_run.write('DMP_CONTEXT.set("{version}", "{contextid}", {data}, {templates});'.format(
            version=__version__,
            contextid=provider_run.uid,
            data=json.dumps(context_data, cls=self.encoder, separators=(',', ':')) if context_data else '{}',
            templates=json.dumps(data['templates']),
        ))
        provider_run.write('</script>')


class jscontext(str):
    '''
    Marks a key in the context dictionary as a JS context item.
    JS context items are sent to the template like normal,
    but they are also added to the runtime JS namespace.

    See the tutorial for more information on this function.
    '''
    # no code needed, just using the class for identity



import os, os.path, sys, re 
from setuptools import setup

MODULE_NAME = 'django_mako_plus'


# I can't import the version file the normal way because it loads
# __init__.py, which then imports the DMP engine.  
with open('django_mako_plus/version.py') as f:
    match = re.search("__version__\s=\s'(\d+\.\d+\.\d+)'", f.read())
    if not match:
        print('Cannot determine the DMP version. Aborting setup.py.')
        sys.exit(1)
    VERSION = match.group(1)


CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Programming Language :: Python :: 3 :: Only',
    'Framework :: Django',
    'Framework :: Django :: 1.11',
    'Framework :: Django :: 1.10',
    'Framework :: Django :: 1.9',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Topic :: Software Development',
]
install_requires = [
    'django >= 1.9.0',
    'mako >= 1.0.0',
]


if len(sys.argv) >= 2 and sys.argv[1] == 'sdist':
    # remove the __pycache__ directories since the ones in project_template seems to stick around
    os.system('find . -name "__pycache__" -type d -exec rm -r "{}" \; > /dev/null')

# Compile the list of packages available
packages = []
def walk(root):
  for fname in os.listdir(root):
    fpath = os.path.join(root, fname)
    # skip hidden/cache files
    if fname.startswith('.') or fname.endswith('.pyc') or fname in ( '__pycache__', 'app_template', 'project_template' ):
      continue
    # if a directory, walk it
    elif os.path.isdir(fpath):
      walk(fpath)
    # if an __init__.py file, add the directory to the packages
    elif fname == '__init__.py':
      packages.append(os.path.dirname(fpath))
walk(MODULE_NAME)

data_files = []
# add the readme/license
data_files.extend([
    ('', [ 'readme.md' ]),
    ('', [ 'readme.txt' ]),
    ('', [ 'license.txt' ]),
])

# add the extra directories
# empty directories within app_template/ will cause problems with distutils, so be sure each directory has at least one file
package_data_files = []
for template_dir in ( 'app_template', 'project_template', 'webroot' ):
    for root, dirs, files in os.walk(os.path.join(MODULE_NAME, template_dir)):
      dirs[:] = [ d for d in dirs if not d.startswith('.') and not d in ( '__pycache__', ) ] # skip hidden and __pycache__ directories
      for fname in files:
        if fname.startswith('.') or fname.endswith('.pyc'): # skip hidden/cache files
          continue
        package_data_files.append(os.path.join(root[len(MODULE_NAME)+1:], fname))

# read the long description if sdist
description = 'Django+Mako: Routing by Convention, Python-Centric Template Language'
long_description = description
if len(sys.argv) > 1 and sys.argv[1] == 'sdist':
  long_description = open('readme.txt').read()

# run the setup
setup(
  name='django-mako-plus',
  description=description,
  long_description=long_description,
  version=VERSION,
  author='Conan Albrecht',
  author_email='doconix@gmail.com',
  url="http://django-mako-plus.readthedocs.io/",
  download_url="https://github.com/doconix/django-mako-plus/archive/master.zip",
#  package_dir={ MODULE_NAME: MODULE_NAME },
  packages=packages,
  package_data = {
    MODULE_NAME: package_data_files,
  },
#  data_files=data_files,
  install_requires=install_requires,
  classifiers=CLASSIFIERS,
  license='Apache 2.0',
)



from django.dispatch import Signal


#############################################################
###   Signals that we send from DMP
###
###   See the standard documentation on Django regarding
###   signals.  Also see the DMP documentation on these
###   specific signals.

#  Triggered just before DMP calls a view's process_request() method.
#  If the method returns an HttpResponse object, processing stops and the object is returned to the browser.
#
#    request     :: The request object, which has several attributes that might be of interest:
#                       request.dmp.app      :: The app name, based on the current url.
#                       request.dmp.page     :: The page name (the views/.py file or templates/.html file), based on the current url.
#                       request.dmp.urlparams           :: Any extra url parameters, based on the current url.
#                       request.dmp.module   :: The module where the view function is located.
#                       request.dmp.function :: The specific view function within the module to call.
#    view_args   :: The list of positional arguments to be sent to the view function.
#    view_kwargs :: The dictionary of keyword arguments to be sent to the view function.
#
dmp_signal_pre_process_request = Signal(providing_args=['request', 'view_args', 'view_kwargs'])

#  Triggered just after a view's process_request() method returns.
#  If the method returns an HttpResponse object, the normal response is replaced with that object.
#
#    request      :: The request object
#    response     :: The return from the process_request method, normally an HttpResponse object.
#    view_args    :: The list of positional arguments that was to the view.
#    view_kwargs  :: The dictionary of keyword arguments that was sent to the view function.
#
dmp_signal_post_process_request = Signal(providing_args=['request', 'response', 'view_args', 'view_kwargs'])

#  Triggered just before DMP renders a Mako template
#  If the method returns a different Template object than the one passed into it, the returned on is used.
#  In other words, this signal lets you override any template just before DMP renders it.
#
#    request  :: the request object
#    context  :: the dict of variables being sent to the template.
#    template :: the Mako template object that will render.
dmp_signal_pre_render_template = Signal(providing_args=['request', 'context', 'template'])

#  Triggered just after DMP renders a Mako template.
#  If the method returns a value, the template-generated content is replaced with that value.  While
#  the template still rendered, its content is discarded and replaced with this return.
#
#    request  :: the request object
#    context  :: the dict of variables being sent to the template.
#    template :: the Mako template object that will render.
#    content  :: the rendered content from the template.
dmp_signal_post_render_template = Signal(providing_args=['request', 'context', 'template', 'content'])

#  Triggered when a RedirectException is encountered in the DMP controller.
#  This signal lets you adjust the values of the exception, such as where it is redirecting to.
#
#    request      :: the request object
#    exc          :: The exception object, including:
#                       exc.redirect_to (new url the router will process with)
#                       exc.permanent   (whether the browser should be told it is a permanent redirect or not)
dmp_signal_redirect_exception = Signal(providing_args=['request', 'exc'])

#  Triggered when an InternalRedirectException is encountered in the DMP controller.
#  This signal lets you adjust the values of the exception, such as where it is redirecting to.
#
#    request      :: the request object
#    exc          :: The exception object, including exc.redirect_to (new url the router will process with).
dmp_signal_internal_redirect_exception = Signal(providing_args=['request', 'exc'])
